@@ -1,28 +1,17 @@
+package Locale::PO;
 use strict;
 use warnings;
-
-package Locale::PO;
+our $VERSION = '0.18';
 
 use Carp;
 
-#use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
-#use locale;
-
-#require Exporter;
-#require AutoLoader;
-
-#@ISA     = qw(Exporter AutoLoader);
-#@EXPORT  = qw();
-our $VERSION = '0.16';
-
-# Preloaded methods go here.
-
 sub new {
     my $this    = shift;
     my %options = @_;
     my $class   = ref($this) || $this;
     my $self    = {};
     bless $self, $class;
+	$self->_flags([]);
     $self->msgid( $options{'-msgid'} ) if defined( $options{'-msgid'} );
     $self->msgid_plural( $options{'-msgid_plural'} )
       if defined( $options{'-msgid_plural'} );
@@ -39,6 +28,7 @@ sub new {
     $self->c_format(1) if defined( $options{'-c_format'} );
     $self->c_format(0) if defined( $options{'-no-c-format'} );
     $self->c_format(0) if defined( $options{'-no_c_format'} );
+	$self->loaded_line_number( $options{'-loaded_line_number'} ) if defined( $options{'-loaded_line_number'} );
     return $self;
 }
 
@@ -98,22 +88,98 @@ sub reference {
     @_ ? $self->{'reference'} = shift: $self->{'reference'};
 }
 
+sub obsolete {
+    my $self = shift;
+    @_ ? $self->{'obsolete'} = shift: $self->{'obsolete'};
+}
+
 sub fuzzy {
     my $self = shift;
-    @_ ? $self->{'fuzzy'} = shift: $self->{'fuzzy'};
+	if (@_) {
+		my $value = shift;
+		$value? $self->add_flag('fuzzy') : $self->remove_flag('fuzzy');
+	}
+    return $self->has_flag('fuzzy');
 }
 
 sub c_format {
     my $self = shift;
-    @_ ? $self->{'c_format'} = shift: $self->{'c_format'};
+	return $self->_tri_value_flag('c-format', @_);
 }
 
 sub php_format {
     my $self = shift;
-    @_ ? $self->{'php_format'} = shift: $self->{'php_format'};
+	return $self->_tri_value_flag('php-format', @_);
 }
 
-sub normalize_str {
+sub _flags {
+    my $self = shift;
+    @_ ? $self->{'_flags'} = shift: $self->{'_flags'};
+}
+
+sub _tri_value_flag {
+    my $self = shift;
+    my $flag_name = shift;
+    if (@_) { # set or clear the flags
+		my $value = shift;
+		if (!defined($value) || $value eq "") {
+			$self->remove_flag("$flag_name");
+			$self->remove_flag("no-$flag_name");
+		    return undef;
+		}
+		elsif ($value) {
+			$self->add_flag("$flag_name");
+			$self->remove_flag("no-$flag_name");
+		    return 1;
+		}
+		else {
+			$self->add_flag("no-$flag_name");
+		    $self->remove_flag("$flag_name");
+		    return 0;
+		}
+    }
+    else { # check the flags
+		return 1 if $self->has_flag("$flag_name");
+		return 0 if $self->has_flag("no-$flag_name");
+		return undef;
+    }
+}
+
+sub add_flag
+{
+	my ($self, $flag_name) = @_;
+	push @{$self->_flags}, $flag_name;
+	return;
+}
+
+sub remove_flag
+{
+	my ($self, $flag_name) = @_;
+	my @new_flags;
+	foreach my $flag (@{$self->_flags})
+	{
+		push @new_flags, $flag unless $flag eq $flag_name;
+	}
+	$self->_flags(\@new_flags);
+	return;
+}
+
+sub has_flag
+{
+	my ($self, $flag_name) = @_;
+	foreach my $flag (@{$self->_flags})
+	{
+		return 1 if $flag eq $flag_name;
+	}
+	return;
+}
+
+sub loaded_line_number {
+    my $self = shift;
+    @_ ? $self->{'loaded_line_number'} = shift: $self->{'loaded_line_number'};
+}
+
+sub _normalize_str {
     my $self     = shift;
     my $string   = shift;
     my $dequoted = $self->dequote($string);
@@ -142,33 +208,30 @@ sub normalize_str {
 
 sub dump {
     my $self = shift;
+	my $obsolete = $self->obsolete? '#~ ' : '';
     my $dump;
-    $dump = $self->dump_multi_comment( $self->comment, "# " )
+    $dump = $self->_dump_multi_comment( $self->comment, "# " )
       if ( $self->comment );
-    $dump .= $self->dump_multi_comment( $self->automatic, "#. " )
+    $dump .= $self->_dump_multi_comment( $self->automatic, "#. " )
       if ( $self->automatic );
-    $dump .= $self->dump_multi_comment( $self->reference, "#: " )
+    $dump .= $self->_dump_multi_comment( $self->reference, "#: " )
       if ( $self->reference );
-    my $flags = '';
-    $flags .= ", fuzzy" if $self->fuzzy;
-    $flags .= ", c-format"
-      if ( defined( $self->c_format )
-        and $self->c_format );
-    $flags .= ", no-c-format"
-      if ( defined( $self->c_format )
-        and !$self->c_format );
-    $flags .= ", php-format"
-      if ( defined( $self->php_format )
-        and $self->php_format );
-    $dump .= "#$flags\n" if length $flags;
-    $dump .= "msgid " . $self->normalize_str( $self->msgid );
-    $dump .= "msgid_plural " . $self->normalize_str( $self->msgid_plural )
-      if $self->msgid_plural;
 
-    $dump .= "msgstr " . $self->normalize_str( $self->msgstr ) if $self->msgstr;
+    my $flags = '';
+	foreach my $flag (@{$self->_flags})
+	{
+		$flags .= ", $flag"
+	}
+    $dump .= "#$flags\n" if length $flags;
+
+    $dump .= "${obsolete}msgid " . $self->_normalize_str( $self->msgid );
+    $dump .= "${obsolete}msgid_plural " . $self->_normalize_str( $self->msgid_plural )
+		if $self->msgid_plural;
+
+    $dump .= "${obsolete}msgstr " . $self->_normalize_str( $self->msgstr ) if $self->msgstr;
 
     if ( my $msgstr_n = $self->msgstr_n ) {
-        $dump .= "msgstr[$_] " . $self->normalize_str( $$msgstr_n{$_} )
+        $dump .= "${obsolete}msgstr[$_] " . $self->_normalize_str( $$msgstr_n{$_} )
           for sort { $a <=> $b } keys %$msgstr_n;
     }
 
@@ -176,7 +239,7 @@ sub dump {
     return $dump;
 }
 
-sub dump_multi_comment {
+sub _dump_multi_comment {
     my $self    = shift;
     my $comment = shift;
     my $leader  = shift;
@@ -207,15 +270,15 @@ sub dequote {
 
 sub save_file_fromarray {
     my $self = shift;
-    $self->save_file( @_, 0 );
+    $self->_save_file( @_, 0 );
 }
 
 sub save_file_fromhash {
     my $self = shift;
-    $self->save_file( @_, 1 );
+    $self->_save_file( @_, 1 );
 }
 
-sub save_file {
+sub _save_file {
     my $self    = shift;
     my $file    = shift;
     my $entries = shift;
@@ -236,19 +299,20 @@ sub save_file {
 
 sub load_file_asarray {
     my $self = shift;
-    $self->load_file( $_[0], 0 );
+    $self->_load_file( $_[0], 0 );
 }
 
 sub load_file_ashash {
     my $self = shift;
-    $self->load_file( $_[0], 1 );
+    $self->_load_file( $_[0], 1 );
 }
 
-sub load_file {
+sub _load_file {
     my $self   = shift;
     my $file   = shift;
     my $ashash = shift;
     my ( @entries, %entries );
+	my $line_number = 0;
     my $po;
     my %buffer;
     my $last_buffer;
@@ -256,6 +320,7 @@ sub load_file {
 
     while (<IN>) {
         chop;
+		$line_number++;
         if (/^$/) {
 
             # Empty line. End of an entry.
@@ -263,8 +328,7 @@ sub load_file {
             if ( defined($po) ) {
 
                 $po->msgid( $buffer{msgid} ) if defined $buffer{msgid};
-                $po->msgid_plural( $buffer{msgid_plural} )
-                  if defined $buffer{msgid_plural};
+                $po->msgid_plural( $buffer{msgid_plural} ) if defined $buffer{msgid_plural};
                 $po->msgstr( $buffer{msgstr} )     if defined $buffer{msgstr};
                 $po->msgstr_n( $buffer{msgstr_n} ) if defined $buffer{msgstr_n};
 
@@ -291,10 +355,10 @@ sub load_file {
                 %buffer = ();
             }
         }
-        elsif ( /^# (.*)/ or /^#()$/ ) {
+        elsif ( /^#\s+(.*)/ or /^#()$/ ) {
 
             # Translator comments
-            $po = new Locale::PO unless defined($po);
+            $po = Locale::PO->new( -loaded_line_number => $line_number ) unless defined($po);
             if ( defined( $po->comment ) ) {
                 $po->comment( $po->comment . "\n$1" );
             }
@@ -302,10 +366,10 @@ sub load_file {
                 $po->comment($1);
             }
         }
-        elsif (/^#\. (.*)/) {
+        elsif (/^#\.\s*(.*)/) {
 
             # Automatic comments
-            $po = new Locale::PO unless defined($po);
+            $po = Locale::PO->new( -loaded_line_number => $line_number ) unless defined($po);
             if ( defined( $po->automatic ) ) {
                 $po->automatic( $po->automatic . "\n$1" );
             }
@@ -313,10 +377,10 @@ sub load_file {
                 $po->automatic($1);
             }
         }
-        elsif (/^#: (.*)/) {
+        elsif (/^#:\s+(.*)/) {
 
             # reference
-            $po = new Locale::PO unless defined($po);
+            $po = Locale::PO->new( -loaded_line_number => $line_number ) unless defined($po);
             if ( defined( $po->reference ) ) {
                 $po->reference( $po->reference . "\n$1" );
             }
@@ -324,45 +388,47 @@ sub load_file {
                 $po->reference($1);
             }
         }
-        elsif (/^#, (.*)/) {
+        elsif (/^#,\s+(.*)/) {
 
             # flags
-            my $flags = $1;
-            $po = new Locale::PO unless defined($po);
-            $po->fuzzy(1)      if $flags =~ /fuzzy/i;
-            $po->c_format(1)   if $flags =~ /c-format/i;
-            $po->c_format(0)   if $flags =~ /no-c-format/i;
-            $po->php_format(1) if $flags =~ /php-format/i;
+            my @flags = split /\s*[,]\s*/, $1;
+            $po = Locale::PO->new( -loaded_line_number => $line_number ) unless defined($po);
+			foreach my $flag (@flags)
+			{
+				$po->add_flag($flag);
+			}
         }
-        elsif (/^msgid (.*)/) {
-            $po = new Locale::PO unless defined($po);
-            $buffer{msgid} = $self->dequote($1);
+        elsif (/^(#~\s+)?msgid\s+(.*)/) {
+            $po = Locale::PO->new( -loaded_line_number => $line_number ) unless defined($po);
+            $buffer{msgid} = $self->dequote($2);
             $last_buffer = \$buffer{msgid};
+			$po->obsolete(1) if $1;
         }
-        elsif (/^msgid_plural (.*)/) {
-            $po = new Locale::PO unless defined($po);
-            $buffer{msgid_plural} = $self->dequote($1);
-            $last_buffer .= \$buffer{msgid_plural};
+        elsif (/^(#~\s+)?msgid_plural\s+(.*)/) {
+            $po = Locale::PO->new( -loaded_line_number => $line_number ) unless defined($po);
+            $buffer{msgid_plural} = $self->dequote($2);
+            $last_buffer = \$buffer{msgid_plural};
+			$po->obsolete(1) if $1;
         }
-        elsif (/^msgstr (.*)/) {
+        elsif (/^(?:#~\s+)?msgstr\s+(.*)/) {
 
             # translated string
             $buffer{msgstr} = $self->dequote($1);
             $last_buffer = \$buffer{msgstr};
         }
-        elsif (/^msgstr\[(\d+)\] (.*)/) {
+        elsif (/^(?:#~\s+)?msgstr\[(\d+)\]\s+(.*)/) {
 
             # translated string
             $buffer{msgstr_n}{$1} = $self->dequote($2);
             $last_buffer = \$buffer{msgstr_n}{$1};
         }
-        elsif (/^"/) {
+        elsif (/^(?:#~\s+)?"/) {
 
             # contined string
             $$last_buffer .= $self->dequote($_);
         }
         else {
-            warn "Strange line in $file: $_\n";
+            warn "Strange line at $file line $line_number: $_\n";
         }
     }
     if ( defined($po) ) {
@@ -371,9 +437,18 @@ sub load_file {
           if defined $buffer{msgid_plural};
         $po->msgstr( $buffer{msgstr} )     if defined $buffer{msgstr};
         $po->msgstr_n( $buffer{msgstr_n} ) if defined $buffer{msgstr_n};
-
-        $entries{ $po->msgid } = $po if $ashash;
-        push( @entries, $po ) unless $ashash;
+		
+		# ashash
+		if ($ashash) {
+			# don't overwrite non-obsolete entries with obsolete ones
+			unless (($po->obsolete) && (defined($entries{ $po->msgid })) && (not $entries{ $po->msgid }->obsolete)) {
+				$entries{ $po->msgid } = $po;
+			}
+		}
+		# asarray
+		else {
+			push( @entries, $po );
+		}
     }
     close IN;
     return ( $ashash ? \%entries : \@entries );
@@ -392,25 +467,25 @@ Locale::PO - Perl module for manipulating .po entries from GNU gettext
 
 =head1 SYNOPSIS
 
-  use Locale::PO;
+	use Locale::PO;
 
-  $po = new Locale::PO([-option=>value,...])
-  [$string =] $po->msgid([new string]);
-  [$string =] $po->msgstr([new string]);
-  [$string =] $po->comment([new string]);
-  [$string =] $po->automatic([new string]);
-  [$string =] $po->reference([new string]);
-  [$value =] $po->fuzzy([value]);
-  [$value =] $po->c_format([value]);
-  print $po->dump;
+	$po = new Locale::PO([-option=>value,...])
+	[$string =] $po->msgid([new string]);
+	[$string =] $po->msgstr([new string]);
+	[$string =] $po->comment([new string]);
+	[$string =] $po->automatic([new string]);
+	[$string =] $po->reference([new string]);
+	[$value =] $po->fuzzy([value]);
+	[$value =] $po->add_flag('c-format');
+	print $po->dump;
 
-  $quoted_string = $po->quote($string);
-  $string = $po->dequote($quoted_string);
+	$quoted_string = $po->quote($string);
+	$string = $po->dequote($quoted_string);
 
-  $aref = Locale::PO->load_file_asarray(<filename>);
-  $href = Locale::PO->load_file_ashash(<filename>);
-  Locale::PO->save_file_fromarray(<filename>,$aref);
-  Locale::PO->save_file_fromhash(<filename>,$href);
+	$aref = Locale::PO->load_file_asarray(<filename>);
+	$href = Locale::PO->load_file_ashash(<filename>);
+	Locale::PO->save_file_fromarray(<filename>,$aref);
+	Locale::PO->save_file_fromhash(<filename>,$href);
 
 =head1 DESCRIPTION
 
@@ -424,11 +499,14 @@ interface in which each entry in a .po file is a Locale::PO object.
 
 =item new
 
+	my Locale::PO $po = new Locale::PO;
+	my Locale::PO $po = new Locale::PO(%options);
+
 Create a new Locale::PO object to represent a po entry.
 You can optionally set the attributes of the entry by passing 
 a list/hash of the form:
 
- -option=>value, -option=>value, etc.
+	-option=>value, -option=>value, etc.
 
 Where options are msgid, msgstr, comment, automatic, reference,
 fuzzy, and c-format. See accessor methods below.
@@ -436,49 +514,74 @@ fuzzy, and c-format. See accessor methods below.
 To generate a po file header, add an entry with an empty
 msgid, like this:
 
-   $po = new Locale::PO(-msgid=>'', -msgstr=>
-	"Project-Id-Version: PACKAGE VERSION\\n" .
-	"PO-Revision-Date: YEAR-MO-DA HO:MI +ZONE\\n" .
-	"Last-Translator: FULL NAME <EMAIL@ADDRESS>\\n" .
-	"Language-Team: LANGUAGE <LL@li.org>\\n" .
-	"MIME-Version: 1.0\\n" .
-	"Content-Type: text/plain; charset=CHARSET\\n" .
-	"Content-Transfer-Encoding: ENCODING\\n");
+	$po = new Locale::PO(-msgid=>'', -msgstr=>
+		"Project-Id-Version: PACKAGE VERSION\\n" .
+		"PO-Revision-Date: YEAR-MO-DA HO:MI +ZONE\\n" .
+		"Last-Translator: FULL NAME <EMAIL@ADDRESS>\\n" .
+		"Language-Team: LANGUAGE <LL@li.org>\\n" .
+		"MIME-Version: 1.0\\n" .
+		"Content-Type: text/plain; charset=CHARSET\\n" .
+		"Content-Transfer-Encoding: ENCODING\\n");
 
 =item msgid
 
 Set or get the untranslated string from the object.
 
+This method expects the new string in unquoted form but returns the current string in quoted form.
+
 =item msgid_plural
 
 Set or get the untranslated plural string from the object.
 
+This method expects the new string in unquoted form but returns the current string in quoted form.
+
 =item msgstr
 
 Set or get the translated string from the object.
 
+This method expects the new string in unquoted form but returns the current string in quoted form.
+
 =item msgstr_n
 
 Get or set the translations if there are purals involved. Takes and
 returns a hashref where the keys are the 'N' case and the values are
 the strings. eg:
 
-    $po->msgstr_n(
-        {
-            0 => 'found %d plural translations',
-            1 => 'found %d singular translation',
-        }
-    );
+	$po->msgstr_n(
+	    {
+	        0 => 'found %d plural translations',
+	        1 => 'found %d singular translation',
+	    }
+	);
+
+This method expects the new strings in unquoted form but returns the current strings in quoted form.
+
+=item obsolete
+
+Returns 1 if the entry is obsolete.
+Obsolete entries have their msgid, msgid_plural, msgstr and msgstr_n lines commented out with "#~"
+
+When using load_file_ashash, non-obsolete entries will always replace obsolete entries with the same msgid.
 
 =item comment
 
 Set or get translator comments from the object.
 
+If there are no such comments, then the value is undef.  Otherwise,
+the value is a string that contains the comment lines delimited with
+"\n".  The string includes neither the S<"# "> at the beginning of
+each comment line nor the newline at the end of the last comment line.
+
 =item automatic
 
 Set or get automatic comments from the object (inserted by 
 emacs po-mode or xgettext).
 
+If there are no such comments, then the value is undef.  Otherwise,
+the value is a string that contains the comment lines delimited with
+"\n".  The string includes neither the S<"#. "> at the beginning of
+each comment line nor the newline at the end of the last comment line.
+
 =item reference
 
 Set or get reference marking comments from the object (inserted
@@ -492,8 +595,41 @@ When setting, use 1 to turn on fuzzy, and 0 to turn it off.
 =item c_format
 
 Set or get the c-format or no-c-format flag on the object.
-This can take 3 values: 1 implies c-format, 0 implies no-c-format,
-and blank or undefined implies neither.
+
+This can take 3 values: 
+1 implies c-format, 0 implies no-c-format, and undefined implies neither.
+
+=item php_format
+
+Set or get the php-format or no-php-format flag on the object.
+
+This can take 3 values:
+1 implies php-format, 0 implies no-php-format, and undefined implies neither.
+
+=item has_flag
+
+	if ($po->has_flag('perl-format')) {
+		...
+	}
+
+Returns true if the flag exists in the entries #~ comment
+
+=item add_flag
+
+	$po->add_flag('perl-format');
+
+Adds the flag to the #~ comment
+
+=item remove_flag
+
+	$po->remove_flag('perl-format');
+
+Removes the flag from the #~ comment
+
+=item loaded_line_number
+
+When using one of the load_file_as* methods,
+this will return the line number that the entry started at in the file.
 
 =item dump
 
@@ -539,13 +675,21 @@ are sorted alphabetically by untranslated string.
 
 =head1 AUTHOR
 
-Alan Schwartz, alansz@pennmush.org
+Ken Prows, perl@xev.net
+
+Original version by: Alan Schwartz, alansz@pennmush.org
 
 =head1 BUGS
 
-If you load_file then save_file, the output file may have slight
+If you load_file_as* then save_file_from*, the output file may have slight
 cosmetic differences from the input file (an extra blank line here or there).
 
+msgid, msgid_plural, msgstr and msgstr_n expect a non-quoted string as input, but return quoted strings.
+I'm hesitant to change this in fear of breaking the modules/scripts of people already using Locale::PO.
+
+Locale::PO requires blank lines between entries, but Uniforum style PO
+files don't have any.
+
 =head1 SEE ALSO
 
 xgettext(1).
